Attribute VB_Name = "Script"
' Basic script system for system verification.
' Loosely based upon the script code from the MonkeyNES emulator;
' but is completely original code and there are some fundamental
' differences in the scripting...
'
' Project: FileStore [E01S emulator]
' Module : Script
'
' Version: 0.00
' Date   : 24th July 2010
' Started: 11th November 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'
'
' COMMAND SUMMARY:
'
'   <reg> = X Y A SP PC PS C Z I D B V N
'                          \-- flags --/
'
'   reset                               - reset cpu
'   memset <addr> [<value>]             - set <addr> to <value> (or zero)
'   memclr [<addr>]                     - zero <addr> (or everything)
'   op <op> [<param1> [<param2>]]       - execute <op> with optional <param...>
'   regset <reg> [<value>]              - set <reg/flag> to <value> (or zero)
'   assert <reg>|m<addr> <cond> <value> - assert <reg> or memory is/not <value>
'                                         <cond> can be '=' (is) or '!' (isn't)
'   load <file>                         - chain <file> (only one at a time!)
'   exec                                - step once from location &0200
'   cont                                - step again (if necessary)
'   msg                                 - report rest of line as message to user
'
' Script files MUST be in <app>\testscripts and have .txt extension.
'
' This is not intended to be a consumer adjustable part, so the parser is very
' simple, few error checks, and it is not terribly obvious how to get this stuff
' even working!
'
' The answer?
' Command line:   FileStore -script <initial file>  [not currently implemented!
'                                                    there's a button on setup dialogue]
'

Option Explicit


Private Const FRMEM As Integer = -2  ' Memory reference
Private Const FRERR As Integer = -1  ' Flag/register error
Private Const XREG  As Integer = 0   ' X
Private Const YREG  As Integer = 1   ' Y
Private Const AREG  As Integer = 2   ' Accumulator
Private Const SPREG As Integer = 3   ' Stack Pointer
Private Const PCREG As Integer = 4   ' Program Counter
Private Const PSREG As Integer = 5   ' Processor Status (as a byte)
Private Const CFLAG As Integer = 6   ' Carry flag
Private Const ZFLAG As Integer = 7   ' Zero flag
Private Const IFLAG As Integer = 8   ' Interrupt disable flag
Private Const DFLAG As Integer = 9   ' Decimal flag
Private Const BFLAG As Integer = 10  ' Break flag
Private Const VFLAG As Integer = 11  ' oVerflow flag
Private Const NFLAG As Integer = 12  ' Negative flag

Private ScriptFile As String         ' initial script file name
Private ScriptLoad As String         ' chained filename
Private ScriptFHdl As Integer        ' script file handle
Private ScriptLHdl As Integer        ' loaded file handle
Private ScriptFNam As String         ' script file name
Private ScriptLNam As String         ' loaded script file name
Private ScriptLine As String         ' line read from script file
Private LastOp As String             ' last op command
Private LineFNum As Integer          ' line number, script file
Private LineLNum As Integer          ' line number, loaded file


Public Sub Script_Initialise(ByVal Filename As String)
  CPU_Initialise
  
  ScriptFile = App.Path & "\testscripts\" & Filename ' relative filename expected
  ScriptFHdl = FreeFile
  Open ScriptFile For Binary Access Read As ScriptFHdl
  ScriptFNam = Filename
  frmSetup.lblTest.Caption = "  Running test script: " & ScriptFNam
  frmSetup.lblTest.Refresh
  DoEvents
  
  LineFNum = 1
  LastOp = ""
  
  ' just go through it...
  While (Loc(ScriptFHdl) < LOF(ScriptFHdl))
  
    If (ScriptLoad <> "") Then
      ' we have a loaded file, so process that
      While (Loc(ScriptLHdl) < LOF(ScriptLHdl))
        Line Input #ScriptLHdl, ScriptLine
        LineLNum = LineLNum + 1
        Debug.Print "L: " & ScriptLine
        Script_ParseCommand
      Wend
      
      Close #ScriptLHdl
      ScriptLoad = ""
      ScriptLNam = ""
      frmSetup.lblTest.Caption = "  Running test script: " & ScriptFNam
      frmSetup.lblTest.Refresh
      DoEvents
      LastOp = 0
      
    Else
      ' no loaded file, process it ourselves
      Line Input #ScriptFHdl, ScriptLine
      LineFNum = LineFNum + 1
      Debug.Print "F: " & ScriptLine
      Script_ParseCommand
    End If
  
  Wend
  Close #ScriptFHdl
  
  ScriptLine = "memclr"
  Script_MemClr
  Script_Reset
  
  MsgBox "Test complete."
End Sub


Private Sub Script_ParseCommand()
  Dim Cmd As String, Posn As Integer, Which As String
  
  Posn = InStr(1, ScriptLine, ";")
  If (Posn <> 0) Then ScriptLine = Trim$(Left$(ScriptLine, (Posn - 1)))
  
  Cmd = LCase$(Script_GetPart(1))
  
  Select Case Cmd
    Case ";", "":
      ' comment line or blank line, ignore
      Exit Sub
      
    Case "reset":
      Script_Reset
      Exit Sub
    
    Case "memset":
      Script_MemSet
      Exit Sub
    
    Case "memclr":
      Script_MemClr
      Exit Sub
    
    Case "op":
      Script_Op
      Exit Sub
        
    Case "regset":
      Script_RegSet
      Exit Sub
    
    Case "assert":
      Script_Assert
      Exit Sub
      
    Case "load":
      Script_Load
      Exit Sub
  
    Case "exec":
      Script_Exec True
      Exit Sub
  
    Case "cont":
      Script_Exec False
      Exit Sub

    Case "msg":
      Script_Msg
      Exit Sub

    Case "diss":
      frmDiss.DoDiss &H200
      frmDiss.Show
      Exit Sub
  
  End Select

  If (ScriptLNam <> "") Then
    Which = "FileStore test script - " & ScriptLNam
  Else
    Which = "FileStore test script - " & ScriptFNam
  End If
  MsgBox "Script command """ & Cmd & """ is not understood.", , Which
End Sub

Private Function Script_GetPart(ByVal Which As Integer) As String
  ' Gets a 'part' of the script command line.
  '
  '   assert X = FF
  '     |    | |  |
  '     |    | |  '-- part 4 "FF"
  '     |    | '----- part 3 "="
  '     |    '------- part 2 "X"
  '     '------------ part 1 "assert"
  '
  ' There must be AT LEAST one space between each 'part'. The following is
  ' not valid:
  '   assert X=FF
  '    ^      ^
  '    part1  part2
  '    okay   "X=FF", WTF?
  
  Dim Posn As Integer, Temp As String
  
  If (ScriptLine = "") Then Exit Function
  
  ' handle first part
  Posn = InStr(1, ScriptLine, " ")
  If (Posn = 0) Then
    If (Which = 1) Then
      Script_GetPart = ScriptLine
    Else
      Script_GetPart = ""
    End If
    Exit Function
  End If
  If (Which = 1) Then
    Script_GetPart = Left$(ScriptLine, (Posn - 1))
    Exit Function
  End If
  
  ' handle second part
  Temp = Mid$(ScriptLine, (Posn + 1))
  Posn = InStr(1, Temp, " ")
  If (Posn = 0) Then
    If (Which = 2) Then
      Script_GetPart = Temp
    Else
      Script_GetPart = ""
    End If
    Exit Function
  End If
  If (Which = 2) Then
    Script_GetPart = Left$(Temp, (Posn - 1))
    Exit Function
  End If
  
  ' handle third part
  Temp = Mid$(Temp, (Posn + 1))
  Posn = InStr(1, Temp, " ")
  If (Posn = 0) Then
    If (Which = 3) Then
      Script_GetPart = Temp
    Else
      Script_GetPart = ""
    End If
    Exit Function
  End If
  If (Which = 3) Then
    Script_GetPart = Left$(Temp, (Posn - 1))
    Exit Function
  End If
  
  ' fourth part is only used in assert, so it's what remains
  Script_GetPart = Mid$(Temp, (Posn + 1))
End Function


Private Function Script_ParseReg(ByVal Which As Integer) As Integer
  ' Parse register/flag (X Y A SP PC PS C Z I D B V N), return reference.
  ' If register begins with 'm', it is a memory reference...
  
  Dim MyTxt As String, Addr As Long
  
  MyTxt = Script_GetPart(Which)
  Script_ParseReg = FRERR
  
  If (LCase$(Left$(MyTxt, 1)) = "m") Then
    Script_ParseReg = FRMEM
    Exit Function
  End If
  
  Select Case UCase$(MyTxt)
    Case "X": Script_ParseReg = XREG
      
    Case "Y": Script_ParseReg = YREG
      
    Case "A": Script_ParseReg = AREG
      
    Case "SP": Script_ParseReg = SPREG
      
    Case "PC": Script_ParseReg = PCREG
      
    Case "PS": Script_ParseReg = PSREG
      
    Case "C": Script_ParseReg = CFLAG
      
    Case "Z": Script_ParseReg = ZFLAG
      
    Case "I": Script_ParseReg = IFLAG
      
    Case "D": Script_ParseReg = DFLAG
      
    Case "B": Script_ParseReg = BFLAG
      
    Case "V": Script_ParseReg = VFLAG
      
    Case "N": Script_ParseReg = NFLAG
  End Select
  
End Function


Private Function Script_ParseAddress(ByVal Which As Integer) As Long
  ' Parse a 16 bit value at line offset <which>, return it; or -1 if missing.
  Dim MyTxt As String
  
  MyTxt = Script_GetPart(Which)
  If (MyTxt = "") Then
    Script_ParseAddress = -1&
  Else
    Script_ParseAddress = CLng("&H" & MyTxt)
  End If
End Function


Private Function Script_ParseValue(ByVal Which As Integer) As Integer
  ' Parse an 8 bit value at line offset <which>, return it; or -1 if missing.
  Dim MyTxt As String
  
  MyTxt = Script_GetPart(Which)
  If (MyTxt = "=") Then
    MyTxt = Script_GetPart((Which + 1))
  End If
  
  If (MyTxt = "") Then
    Script_ParseValue = 0
  Else
    Script_ParseValue = CInt("&H" & MyTxt)
  End If
End Function


Private Sub Script_Reset()
  CPU_Reset
  Memory_Read &HFC08& ' disable EPROM latch so we're in RAM
End Sub


Private Sub Script_MemSet()
  Dim Addr As Long, Value As Integer
  
  Addr = Script_ParseAddress(2)
  Value = Script_ParseValue(3)
  If (Value = -1) Then Value = 0

  Memory_Write Value, Addr
End Sub


Private Sub Script_MemClr()
  Dim Addr As Long
  
  Addr = Script_ParseAddress(2)
  If (Addr <> -1) Then
    Memory_Write 0, Addr
  Else
    For Addr = &H0& To &HFBFF&
      Memory_Write 0, Addr
    Next
    ' miss out page &FCxx
    For Addr = &HFD00& To &HFFFF&
      Memory_Write 0, Addr
    Next
  End If
End Sub


Private Sub Script_Op()
  Dim Op As Integer, Param1 As Integer, Param2 As Integer, Chnge As Integer
  
  Op = Script_ParseValue(2)
  Param1 = Script_ParseValue(3)
  If (Param1 > 255) Then
    MsgBox "Value greater than &FF specified, will try to correct." & vbCrLf & _
           vbCrLf & "Input: " & ScriptLine & vbCrLf & vbCrLf & "You should " & _
           "use ""op <op> <lowbyte> <highbyte>"" instead of ""op <op> <word>"".", _
           vbExclamation
    Param2 = Param1 \ 256
    Param1 = Param1 And 255
  Else
    If (Param1 = -1) Then Param1 = 0
    Param2 = Script_ParseValue(4)
    If (Param2 = -1) Then Param2 = 0
  End If
  Memory_Write Op, &H200&
  Memory_Write Param1, &H201&
  Memory_Write Param2, &H202&
  CPU_PrepForTest
  CPU_Execute False
  
  LastOp = Trim$(ScriptLine)
End Sub


Private Sub Script_RegSet()
  Dim Reg As Long, Value As Integer, Which As String
  
  If (ScriptLNam <> "") Then
    Which = "FileStore test script - " & ScriptLNam
  Else
    Which = "FileStore test script - " & ScriptFNam
  End If
  
  Reg = Script_ParseReg(2)
  If (Reg = FRMEM) Then
    MsgBox "Can't regset MEMORY, use memset!", , Which
    Exit Sub
  End If
  Value = Script_ParseValue(3)
  If (Value = -1) Then Value = 0
  
  ' NOTE:
  '   We provide <value> as a hex string in order to work with
  '   earlier code, in preference to two functions that would
  '   do essentially the exact same thing...
  
  Select Case Reg
    Case XREG:
      CPU_SetReg 0, Hex(Value)
      
    Case YREG:
      CPU_SetReg 1, Hex(Value)
      
    Case AREG:
      CPU_SetReg 2, Hex(Value)
      
    Case SPREG:
      CPU_SetReg -3, Hex(Value) ' secret value to SP without prompt
      
    Case PCREG:
      MsgBox "Can't set PC directly like this.", , Which
      
    Case PSREG:
      MsgBox "Can't set PS directly like this.", , Which
  
    Case CFLAG:
      CPU_SetFlag 3, IIf((Value <> 0), True, False)
    
    Case ZFLAG:
      CPU_SetFlag 1, IIf((Value <> 0), True, False)
    
    Case IFLAG:
      CPU_SetFlag 6, IIf((Value <> 0), True, False)
    
    Case DFLAG:
      CPU_SetFlag 5, IIf((Value <> 0), True, False)
  
    Case BFLAG:
      CPU_SetFlag 4, IIf((Value <> 0), True, False)
  
    Case VFLAG:
      CPU_SetFlag 2, IIf((Value <> 0), True, False)
  
    Case NFLAG:
      CPU_SetFlag 0, IIf((Value <> 0), True, False)
  
  End Select
End Sub


Private Sub Script_Assert()
  ' assert <reg>|m<addr> <cond> <value>

  Dim Reg As Integer, Cond As String, Value As Long, Addr As Long
  Dim Temp As String, EveVal As Long, Which As String, Suffix As String

  If (ScriptLNam <> "") Then
    Which = "FileStore test script - " & ScriptLNam
  Else
    Which = "FileStore test script - " & ScriptFNam
  End If

  Debug.Print ScriptLine

  Reg = Script_ParseReg(2)
  Cond = Script_GetPart(3)
  Value = Script_ParseAddress(4)

  If (Reg = FRMEM) Then
    ' Handle memory
    Temp = Script_GetPart(2)
    Addr = CLng("&H" & Mid$(Temp, 2))
    EveVal = Memory_Read(Addr)
    
    Debug.Print "  mem check - addr &" & Hex(Addr) & " (is &" & _
                Hex(EveVal) & " vs value &" & Hex(Value) & ")"
  Else
    ' Handle register
    If (Reg = PCREG) Then
      Cycles = 0
      EveVal = CPU_ReadPC
    ElseIf (Reg = PSREG) Then
      ' build it
      EveVal = CPU_AssembleStatus
    Else
      EveVal = CPU_ReadReg(Reg)
    End If
  
    Debug.Print "  mem check - reg #" & CStr(Reg) & " (is &" & _
                Hex(EveVal) & " vs value &" & Hex(Value) & ")"
  End If
  
  Suffix = vbCrLf & vbCrLf
  If (ScriptLoad <> "") Then
    Suffix = Suffix & "At line " & CStr(LineLNum) & " in lscript." & vbCrLf & vbCrLf
  Else
    Suffix = Suffix & "At line " & CStr(LineFNum) & " in fscript." & vbCrLf & vbCrLf
  End If
  If (LastOp <> "") Then
    Suffix = Suffix & "Last op (may or may not be useful) was:" & vbCrLf & _
             vbTab & LastOp
  End If
  
  If (Cond = "=") Then
    If (EveVal <> Value) Then
      MsgBox "Assert equals failed:" & vbCrLf & vbCrLf & ScriptLine & Suffix, , Which
    Else
      Debug.Print "  - tested okay"
    End If
  
  ElseIf (Cond = "!") Then
    If (EveVal = Value) Then
      MsgBox "Assert not-equals failed:" & vbCrLf & vbCrLf & ScriptLine & Suffix, , Which
    Else
      Debug.Print "  - tested okay"
    End If
  
  Else
    MsgBox "Assert comparison """ & Cond & """ unknown, should be " & _
           """="" (equals) or ""!"" (not equals).", , Which
  End If
End Sub


Private Sub Script_Load()
  ' load <file>
  Dim MyScript As String
  
  If (ScriptLoad <> "") Then
    MsgBox "It is only possible to chain once; you are trying to chain from a " & _
           "chained script. In order to do this, you'll need to create a new " & _
           "script that simply chains all the desired scripts in turn...", vbExclamation
    Exit Sub
  End If
  
  LineLNum = 0
  
  MyScript = Script_GetPart(2)
  
  ScriptLoad = App.Path & "\testscripts\" & MyScript
  If (Dir(ScriptLoad) = "") Then
    ScriptLoad = ""
    MsgBox "Script """ & MyScript & """ not found, skipping.", vbExclamation
    Exit Sub
  End If
  
  ScriptLNam = MyScript
  frmSetup.lblTest.Caption = "  Running test script: " & ScriptLNam
  frmSetup.lblTest.Refresh
  DoEvents

  ScriptLHdl = FreeFile
  Open ScriptLoad For Binary Access Read As ScriptLHdl
  
  Debug.Print "### Chaining into " & MyScript & " ###"
End Sub


Private Sub Script_Exec(IsInit As Boolean)
  If (IsInit = True) Then CPU_PrepForTest
  CPU_Execute False
End Sub


Private Sub Script_Msg()
  Dim MyScript As String
  
  MyScript = Mid$(Trim$(ScriptLine), 5)
  MsgBox MyScript
End Sub
